package broker_test

import (
	"encoding/json"
	"fmt"

	. "github.com/onsi/ginkgo"
	. "github.com/onsi/gomega"
	"github.com/pivotal-cf/brokerapi"
	"github.com/pivotal-cf/on-demand-service-broker/broker"
)

var _ = Describe("provisioning", func() {
	var (
		planID string

		serviceSpec  brokerapi.ProvisionedServiceSpec
		provisionErr error

		manifest         = []byte("a-manifest")
		serviceID        = "a-service-id"
		organizationGUID = "a-cf-org"
		spaceGUID        = "a-cf-space"
		jsonParams       []byte
		arbParams        = map[string]interface{}{"foo": "bar"}

		asyncAllowed = true
	)

	BeforeEach(func() {
		planID = existingPlanID
		asyncAllowed = true
		serviceAdapter.GenerateManifestReturns(manifest, nil)
		instanceRepo.InstanceExistsReturns(false, nil)
		var err error
		jsonParams, err = json.Marshal(arbParams)
		Expect(err).NotTo(HaveOccurred())
	})

	JustBeforeEach(func() {
		serviceSpec, provisionErr = b.Provision("some-instance", brokerapi.ProvisionDetails{
			PlanID:           planID,
			RawParameters:    jsonParams,
			OrganizationGUID: organizationGUID,
			SpaceGUID:        spaceGUID,
			ServiceID:        serviceID,
		}, asyncAllowed)
	})

	Context("when bosh deploys the release successfully", func() {
		BeforeEach(func() {
			instanceRepo.SaveReturns(nil)
		})

		It("does not error", func() {
			Expect(provisionErr).NotTo(HaveOccurred())
		})

		It("reports that the provisioning was asynchronous", func() {
			Expect(serviceSpec.IsAsync).To(BeTrue())
		})

		It("invokes the deployer", func() {
			Expect(fakeDeployer.DeployCallCount()).To(Equal(1))
			actualInstance, actualPlan, actualDeploymentName, actualOperationType, actualRequestParams, actualOldBostManifest, actualOldPlanID := fakeDeployer.DeployArgsForCall(0)
			Expect(actualOldBostManifest).To(BeNil())
			Expect(actualRequestParams).To(Equal(map[string]interface{}{
				"plan_id":           planID,
				"parameters":        arbParams,
				"organization_guid": organizationGUID,
				"space_guid":        spaceGUID,
				"service_id":        serviceID,
			}))
			Expect(actualPlan).To(Equal(planID))
			Expect(actualInstance).To(Equal(broker.Instance{ID: "some-instance", PlanID: planID}))
			Expect(actualDeploymentName).To(Equal(broker.InstancePrefix + "some-instance"))
			Expect(actualOperationType).To(Equal(broker.Create))
			Expect(actualOldPlanID).To(BeNil())
		})
	})

	Context("when invalid json params are provided by the broker api", func() {
		BeforeEach(func() {
			jsonParams = []byte("not a valid json")
		})

		It("wraps the returns a raw params invalid error", func() {
			Expect(provisionErr).To(Equal(brokerapi.ErrRawParamsInvalid))
		})
	})

	Context("when no arbitrary params are passed by user", func() {
		BeforeEach(func() {
			jsonParams = []byte{}
		})

		It("return no error", func() {
			Expect(provisionErr).NotTo(HaveOccurred())
		})

		It("no arbitrary params are passed to the adapter", func() {
			_, _, _, _, actualRequestParams, _, _ := fakeDeployer.DeployArgsForCall(0)
			Expect(actualRequestParams["parameters"]).To(BeNil())
		})

		It("invokes the deployer", func() {
			// don't need to fully replicate the conditions above, we just want to ensure that deployment still happens
			// alternatively, use shared example?
			Expect(fakeDeployer.DeployCallCount()).To(Equal(1))
		})
	})

	Context("when a deployment has an internal error", func() {
		BeforeEach(func() {
			fakeDeployer.DeployReturns(broker.Instance{}, fmt.Errorf("fooo"))
		})

		It("logs the error", func() {
			Expect(logBuffer.String()).To(ContainSubstring("error: fooo"))
		})

		It("returns a standard error", func() {
			Expect(provisionErr).To(MatchError(ContainSubstring("An internal error has occurred")))
		})
	})

	Context("when a deployment has a user displayable error", func() {
		BeforeEach(func() {
			fakeDeployer.DeployReturns(broker.Instance{}, broker.DisplayableError{fmt.Errorf("fooo")})
		})

		It("returns the error", func() {
			Expect(provisionErr).To(MatchError(ContainSubstring("fooo")))
		})
	})

	Context("when a provision of an already provisioned instance is triggered", func() {
		BeforeEach(func() {
			instanceRepo.InstanceExistsReturns(true, nil)
		})
		It("returns an error", func() {
			Expect(provisionErr).To(Equal(brokerapi.ErrInstanceAlreadyExists))
		})
	})

	Context("when the async allowed flag is false", func() {
		BeforeEach(func() {
			asyncAllowed = false
		})
		It("returns a  error", func() {
			Expect(provisionErr).To(Equal(brokerapi.ErrAsyncRequired))
		})
	})

	Context("when the count by plan provisioner id fails", func() {
		BeforeEach(func() {
			instanceRepo.CountByPlanIDReturns(0, fmt.Errorf("count fail"))
		})
		It("returns a standard error", func() {
			Expect(provisionErr).To(MatchError(ContainSubstring("An internal error has occurred")))
		})

		It("logs the error", func() {
			Expect(logBuffer.String()).To(ContainSubstring("count fail"))
		})

		It("makes no deployments", func() {
			Expect(fakeDeployer.DeployCallCount()).To(BeZero())
		})
	})

	Context("when the plan quota is reached", func() {
		BeforeEach(func() {
			instanceRepo.CountByPlanIDReturns(serviceInstanceLimit, nil)
		})

		It("returns an error", func() {
			Expect(provisionErr).To(HaveOccurred())
		})

		It("makes no deployments", func() {
			Expect(fakeDeployer.DeployCallCount()).To(BeZero())
		})
	})

	Context("when no quota is specified", func() {
		BeforeEach(func() {
			planID = secondPlanID
		})

		It("succeeds", func() {
			Expect(provisionErr).NotTo(HaveOccurred())
		})

		It("makes a deployment", func() {
			Expect(fakeDeployer.DeployCallCount()).To(Equal(1))
		})
	})

	Context("when plan id given is not configured", func() {
		BeforeEach(func() {
			planID = "non-existant-pland"
		})

		It("return an error", func() {
			Expect(provisionErr).To(HaveOccurred())
		})
	})

})
