package external_test

import (
	"io/ioutil"
	"os"
	"os/exec"

	. "github.com/onsi/ginkgo"
	. "github.com/onsi/gomega"

	"github.com/pivotal-cf/on-demand-service-broker/external"
)

var _ = Describe("CommandRunner", func() {
	var (
		scriptPath string

		stdout string
		stderr string
		runErr error
	)

	JustBeforeEach(func() {
		cmd := exec.Command(scriptPath)

		runner := external.NewCommandRunner()
		var stdoutBytes, stderrBytes []byte
		stdoutBytes, stderrBytes, runErr = runner.Run(cmd)
		stdout = string(stdoutBytes)
		stderr = string(stderrBytes)
	})

	var createScript = func(script string) string {
		tempFile, err := ioutil.TempFile("", "cmd")
		Expect(err).NotTo(HaveOccurred())

		_, err = tempFile.WriteString("#! /bin/sh\n")
		Expect(err).NotTo(HaveOccurred())
		_, err = tempFile.WriteString(script)
		Expect(err).NotTo(HaveOccurred())

		err = tempFile.Chmod(0755)
		Expect(err).NotTo(HaveOccurred())

		err = tempFile.Close()
		Expect(err).NotTo(HaveOccurred())

		return tempFile.Name()
	}

	AfterEach(func() {
		os.Remove(scriptPath)
	})

	Context("when the command runs normally", func() {
		BeforeEach(func() {
			scriptPath = createScript("echo output; echo error >&2")
		})

		It("returns the text written to standard output", func() {
			Expect(stdout).To(Equal("output\n"))
		})

		It("returns the text written to standard error", func() {
			Expect(stderr).To(Equal("error\n"))
		})

		It("returns no error", func() {
			Expect(runErr).NotTo(HaveOccurred())
		})
	})

	Context("when the command does not exist", func() {
		BeforeEach(func() {
			scriptPath = "/no/such/script"
		})

		It("returns an error", func() {
			Expect(runErr).To(HaveOccurred())
		})
	})

	Context("when the command exists but is not executable", func() {
		BeforeEach(func() {
			scriptPath = createScript("this is not a script")
			os.Chmod(scriptPath, 0644)
		})

		It("returns an error", func() {
			Expect(runErr).To(HaveOccurred())
		})
	})

	Context("when the command exits with nonzero status", func() {
		BeforeEach(func() {
			scriptPath = createScript("echo output; echo error >&2; exit 23")
		})

		It("returns the text written to standard output", func() {
			Expect(stdout).To(Equal("output\n"))
		})

		It("returns the text written to standard error", func() {
			Expect(stderr).To(Equal("error\n"))
		})

		It("returns an error", func() {
			Expect(runErr).To(HaveOccurred())
		})
	})
})
