package external

import (
	"encoding/json"
	"fmt"
	"log"
	"os/exec"

	"github.com/pivotal-cf/on-demand-service-broker-sdk/bosh"
	"github.com/pivotal-cf/on-demand-service-broker-sdk/serviceadapter"
)

type Adapter struct {
	ExternalBinPath string
	CommandRunner   CommandRunner
	Logger          *log.Logger
}

func (a *Adapter) GenerateManifest(serviceDeployment serviceadapter.ServiceDeployment, plan serviceadapter.Plan, requestParams map[string]interface{}, previousManifest []byte, previousPlan *serviceadapter.Plan) ([]byte, error) {
	errs := func(err error) ([]byte, error) {
		return nil, err
	}

	serialisedServiceDeployment, err := json.Marshal(serviceDeployment)
	if err != nil {
		return errs(err)
	}

	plan.Properties = SanitiseForJSON(plan.Properties)
	serialisedPlan, err := json.Marshal(plan)
	if err != nil {
		return errs(err)
	}

	serialisedRequestParams, err := json.Marshal(requestParams)
	if err != nil {
		return errs(err)
	}
	if previousPlan != nil {
		previousPlan.Properties = SanitiseForJSON(previousPlan.Properties)
	}

	serialisedPreviousPlan, err := json.Marshal(previousPlan)
	if err != nil {
		return errs(err)
	}

	cmd := exec.Command(
		a.ExternalBinPath, "generate-manifest", string(serialisedServiceDeployment),
		string(serialisedPlan), string(serialisedRequestParams), string(previousManifest), string(serialisedPreviousPlan),
	)
	stdout, stderr, err := a.CommandRunner.Run(cmd)

	if err != nil {
		return errs(fmt.Errorf("an error occurred running external service adapter at %s: '%s'. Stdout: '%s', stderr: '%s'", a.ExternalBinPath, err, string(stdout), string(stderr)))
	}

	a.Logger.Printf("service adapter ran generate-manifest successfully, stderr logs: %s", string(stderr))

	return stdout, nil
}

func SanitiseForJSON(properties serviceadapter.Properties) serviceadapter.Properties {
	propertiesToReturn := serviceadapter.Properties{}

	for k, v := range properties {
		propertiesToReturn[k] = sanitiseValueForJSON(v)
	}

	return propertiesToReturn
}

// structures parsed from YAML contain map[interface{}]interface{}, but JSON marshalling requires map[string]interface{}, so we have to walk the structure and convert any instances of the former to the latter
func sanitiseValueForJSON(obj interface{}) interface{} {
	switch castObj := obj.(type) {

	case map[interface{}]interface{}:
		sanitisedMap := map[string]interface{}{}
		for key, value := range castObj {
			sanitisedMap[key.(string)] = sanitiseValueForJSON(value)
		}
		return sanitisedMap

	case []interface{}:
		var sanitisedSlice []interface{}
		for _, value := range castObj {
			sanitisedSlice = append(sanitisedSlice, sanitiseValueForJSON(value))
		}
		return sanitisedSlice

	default:
		return obj

	}
}

func (a *Adapter) CreateBinding(bindingID string, deploymentTopology bosh.BoshVMs, manifest []byte, arbitraryParams map[string]interface{}) (map[string]interface{}, error) {
	errs := func(err error) (map[string]interface{}, error) {
		return nil, err
	}

	serialisedBoshVMs, err := json.Marshal(deploymentTopology)
	if err != nil {
		return errs(err)
	}
	serialisedArbitraryParams, err := json.Marshal(arbitraryParams)
	if err != nil {
		return errs(err)
	}

	cmd := exec.Command(a.ExternalBinPath, "create-binding", bindingID, string(serialisedBoshVMs), string(manifest), string(serialisedArbitraryParams))
	stdout, stderr, err := a.CommandRunner.Run(cmd)
	if err != nil {
		return errs(fmt.Errorf("an error occurred running external service adapter at %s: '%s'. Stdout: '%s', stderr: '%s'", a.ExternalBinPath, err, string(stdout), string(stderr)))
	}

	a.Logger.Printf("service adapter ran create-binding successfully, stderr logs: %s", string(stderr))

	var credentials map[string]interface{}
	if err := json.Unmarshal(stdout, &credentials); err != nil {
		return errs(err)
	}

	return credentials, nil
}

func (a *Adapter) DeleteBinding(bindingID string, deploymentTopology bosh.BoshVMs, manifest []byte) error {

	serialisedBoshVMs, err := json.Marshal(deploymentTopology)
	if err != nil {
		return err
	}

	cmd := exec.Command(a.ExternalBinPath, "delete-binding", bindingID, string(serialisedBoshVMs), string(manifest))
	stdout, stderr, err := a.CommandRunner.Run(cmd)

	if err != nil {
		return fmt.Errorf("an error occurred running external service adapter at %s: '%s'. Stdout: '%s', stderr: '%s'", a.ExternalBinPath, err, string(stdout), string(stderr))
	}

	return nil
}
