package external_test

import (
	"encoding/json"
	"errors"
	"log"

	. "github.com/onsi/ginkgo"
	. "github.com/onsi/gomega"
	"github.com/pivotal-cf/on-demand-service-broker-sdk/bosh"
	"github.com/pivotal-cf/on-demand-service-broker-sdk/serviceadapter"
	"github.com/pivotal-cf/on-demand-service-broker/external"
	"github.com/pivotal-cf/on-demand-service-broker/external/fake_command_runner"
)

var _ = Describe("external service adapter", func() {
	const externalBinPath = "/thing"
	const manifestContent = "this is a genuine manifest"

	var (
		a         *external.Adapter
		cmdRunner *fake_command_runner.FakeCommandRunner
	)

	BeforeEach(func() {
		cmdRunner = new(fake_command_runner.FakeCommandRunner)
		a = &external.Adapter{
			CommandRunner:   cmdRunner,
			ExternalBinPath: externalBinPath,
			Logger:          log.New(GinkgoWriter, "[unit-tests] ", log.LstdFlags),
		}
	})

	Describe("generating manifest", func() {
		var (
			serviceDeployment serviceadapter.ServiceDeployment
			plan              serviceadapter.Plan
			previousPlan      *serviceadapter.Plan
			params            map[string]interface{}
			previousManifest  []byte

			manifest    []byte
			generateErr error
		)

		BeforeEach(func() {
			cmdRunner.RunReturns([]byte(manifestContent), []byte(""), nil)

			serviceDeployment = serviceadapter.ServiceDeployment{
				DeploymentName: "a-service-deployment",
				Releases: serviceadapter.ServiceReleases{
					{Name: "a-bosh-release"},
				},
				Stemcell: serviceadapter.Stemcell{
					OS:      "BeOS",
					Version: "2",
				},
			}

			plan = serviceadapter.Plan{
				Properties: serviceadapter.Properties{
					"foo": "bar",
					"baz": map[interface{}]interface{}{
						"qux": "quux",
					},
				},
			}
			params = map[string]interface{}{
				"key": "value",
				"anotherkey": map[string]interface{}{
					"innerkey": "innervalue",
				},
			}
			previousManifest = []byte("a-manifest")
			previousPlan = &serviceadapter.Plan{
				Properties: serviceadapter.Properties{
					"previous": "props",
					"baz": map[interface{}]interface{}{
						"qux": "quux",
					},
				},
			}
		})

		JustBeforeEach(func() {
			manifest, generateErr = a.GenerateManifest(serviceDeployment, plan, params, previousManifest, previousPlan)
		})

		It("invokes external manifest generator with serialised parameters", func() {
			serialisedServiceDeployment, err := json.Marshal(serviceDeployment)
			Expect(err).NotTo(HaveOccurred())

			plan.Properties = external.SanitiseForJSON(plan.Properties)
			serialisedPlan, err := json.Marshal(plan)
			Expect(err).NotTo(HaveOccurred())

			serialisedParams, err := json.Marshal(params)
			Expect(err).NotTo(HaveOccurred())

			serialisedPreviousPlan, err := json.Marshal(previousPlan)
			Expect(err).NotTo(HaveOccurred())

			Expect(cmdRunner.RunCallCount()).To(Equal(1))
			argsPassed := cmdRunner.RunArgsForCall(0).Args
			Expect(argsPassed).To(ConsistOf(externalBinPath, "generate-manifest",
				string(serialisedServiceDeployment), string(serialisedPlan),
				string(serialisedParams), string(previousManifest), string(serialisedPreviousPlan)))
		})

		Context("when the external service adapter succeeds", func() {
			It("returns no error", func() {
				Expect(generateErr).ToNot(HaveOccurred())
			})

			It("returns the deserialised stdout from the external process as a bosh manifest", func() {
				Expect(manifest).To(Equal([]byte(manifestContent)))
			})
		})

		Context("when the external service adapter fails", func() {
			BeforeEach(func() {
				cmdRunner.RunReturns([]byte("I'm stdout"), []byte("I'm stderr"), errors.New("oops"))
			})

			It("returns an error", func() {
				Expect(generateErr).To(MatchError("an error occurred running external service adapter at /thing: 'oops'. Stdout: 'I'm stdout', stderr: 'I'm stderr'"))
			})
		})

		Context("previous plan is nil", func() {
			BeforeEach(func() {
				previousPlan = nil
			})

			It("returns no error", func() {
				Expect(generateErr).ToNot(HaveOccurred())
			})

			It("it writes 'null' to the argument list", func() {
				argsPassed := cmdRunner.RunArgsForCall(0).Args
				Expect(argsPassed[6]).To(Equal("null"))
			})
		})
	})

	Describe("creating binding", func() {

		var (
			bindingID          string
			deploymentTopology bosh.BoshVMs
			manifest           []byte
			arbitraryParams    map[string]interface{}

			credentials      map[string]interface{}
			createBindingErr error
		)

		BeforeEach(func() {
			cmdRunner.RunReturns([]byte(`{"username":"bob"}`), []byte(""), nil)

			bindingID = "the-binding"
			deploymentTopology = bosh.BoshVMs{"the-deployment": []string{"a-vm"}}
			manifest = []byte("a-manifest")
			arbitraryParams = map[string]interface{}{"foo": "bar"}
		})

		JustBeforeEach(func() {
			credentials, createBindingErr = a.CreateBinding(bindingID, deploymentTopology, manifest, arbitraryParams)
		})

		It("invokes external binding creator with serialised params", func() {
			serialisedVMs, err := json.Marshal(deploymentTopology)
			Expect(err).NotTo(HaveOccurred())

			serialisedArbitraryParams, err := json.Marshal(arbitraryParams)
			Expect(err).NotTo(HaveOccurred())

			Expect(cmdRunner.RunCallCount()).To(Equal(1))
			argsPassed := cmdRunner.RunArgsForCall(0).Args
			Expect(argsPassed).To(ConsistOf(externalBinPath, "create-binding", bindingID, string(serialisedVMs), string(manifest), string(serialisedArbitraryParams)))
		})

		Context("when the external adapter succeeds", func() {
			It("returns no error", func() {
				Expect(createBindingErr).ToNot(HaveOccurred())
			})

			It("returns the service-specific binding output", func() {
				Expect(credentials).To(Equal(map[string]interface{}{"username": "bob"}))
			})
		})

		Context("when the external service adapter fails", func() {
			BeforeEach(func() {
				cmdRunner.RunReturns([]byte("I'm stdout"), []byte("I'm stderr"), errors.New("oops"))
			})

			It("returns an error", func() {
				Expect(createBindingErr).To(MatchError("an error occurred running external service adapter at /thing: 'oops'. Stdout: 'I'm stdout', stderr: 'I'm stderr'"))
			})
		})
	})

	Describe("deleting binding", func() {
		var (
			bindingID          string
			deploymentTopology bosh.BoshVMs
			manifest           []byte

			deleteBindingError error
		)

		BeforeEach(func() {
			cmdRunner.RunReturns([]byte(""), []byte(""), nil)

			bindingID = "the-binding"
			deploymentTopology = bosh.BoshVMs{}
			manifest = []byte("a-manifest")
		})

		JustBeforeEach(func() {
			deleteBindingError = a.DeleteBinding(bindingID, deploymentTopology, manifest)
		})

		It("invokes external binary with params to delete binding", func() {
			serialisedBoshVMs, err := json.Marshal(deploymentTopology)
			Expect(err).NotTo(HaveOccurred())

			Expect(cmdRunner.RunCallCount()).To(Equal(1))
			argsPassed := cmdRunner.RunArgsForCall(0).Args
			Expect(argsPassed).To(ConsistOf(externalBinPath, "delete-binding", bindingID, string(serialisedBoshVMs), string(manifest)))
		})

		Context("when the external adapter succeeds", func() {
			It("returns no error", func() {
				Expect(deleteBindingError).ToNot(HaveOccurred())
			})
		})

		Context("when the external adapter fails", func() {
			BeforeEach(func() {
				cmdRunner.RunReturns([]byte("I'm stdout"), []byte("I'm stderr"), errors.New("oops"))
			})

			It("returns an error", func() {
				Expect(deleteBindingError).To(MatchError("an error occurred running external service adapter at /thing: 'oops'. Stdout: 'I'm stdout', stderr: 'I'm stderr'"))
			})
		})
	})
})
